@@ -1,6 +1,6 @@
 // Copyright 2008-present Contributors to the OpenImageIO project.
 // SPDX-License-Identifier: BSD-3-Clause
-// https://github.com/OpenImageIO/oiio/blob/master/LICENSE.md
+// https://github.com/OpenImageIO/oiio
 
 /*
   A few bits here are based upon code from NVIDIA that was also released
@@ -8,7 +8,7 @@
      Copyright 2004 NVIDIA Corporation. All Rights Reserved.
 
   Some parts of this file were first open-sourced in Open Shading Language,
-  also 3-Clause BSD licesne, then later moved here. The original copyright
+  also 3-Clause BSD license, then later moved here. The original copyright
   notice was:
      Copyright (c) 2009-2014 Sony Pictures Imageworks Inc., et al.
 
@@ -37,6 +37,7 @@
 #include <typeinfo>
 #include <type_traits>
 
+#include <OpenImageIO/Imath.h>
 #include <OpenImageIO/span.h>
 #include <OpenImageIO/dassert.h>
 #include <OpenImageIO/oiioversion.h>
@@ -192,12 +193,15 @@ inline OIIO_HOSTDEVICE int pow2rounddown(int x) { return floor2(x); }
 
 
 
-/// Round value up to the next whole multiple.
-/// For example, round_to_multiple(7,10) returns 10.
+/// Round value up to the next whole multiple. For example,
+/// `round_to_multiple(10,10) == 10`, `round_to_multiple(17,10) == 20`, and
+/// `round_to_multiple(-17,10) == -10`.
 template <typename V, typename M>
 inline OIIO_HOSTDEVICE V round_to_multiple (V value, M multiple)
 {
-    return V (((value + V(multiple) - 1) / V(multiple)) * V(multiple));
+    if (value >= 0)
+        value += V(multiple) - 1;
+    return value - (value % V(multiple));
 }
 
 
@@ -216,6 +220,20 @@ round_to_multiple_of_pow2(T x, T m)
 
 
 
+/// Round `value` down to a whole multiple of `multiple`. For example,
+/// `round_down_to_multiple(10,10) == 10`,
+/// `round_down_to_multiple(17,10) == 10`, and
+/// `round_down_to_multiple(-17,10) == -20`.
+template <typename V, typename M>
+inline OIIO_HOSTDEVICE V round_down_to_multiple (V value, M multiple)
+{
+    if (value < 0)
+        value -= V(multiple - 1);
+    return value - (value % V(multiple));
+}
+
+
+
 /// Multiply two unsigned 32-bit ints safely, carefully checking for
 /// overflow, and clamping to uint32_t's maximum value.
 inline OIIO_HOSTDEVICE uint32_t
@@ -361,6 +379,25 @@ clamp (const simd::vfloat16& a, const simd::vfloat16& low, const simd::vfloat16&
 {
     return simd::min (high, simd::max (low, a));
 }
+
+// Specialization of clamp for vint4
+template<> OIIO_FORCEINLINE simd::vint4
+clamp (const simd::vint4& a, const simd::vint4& low, const simd::vint4& high)
+{
+    return simd::min (high, simd::max (low, a));
+}
+
+template<> OIIO_FORCEINLINE simd::vint8
+clamp (const simd::vint8& a, const simd::vint8& low, const simd::vint8& high)
+{
+    return simd::min (high, simd::max (low, a));
+}
+
+template<> OIIO_FORCEINLINE simd::vint16
+clamp (const simd::vint16& a, const simd::vint16& low, const simd::vint16& high)
+{
+    return simd::min (high, simd::max (low, a));
+}
 #endif
 
 
@@ -714,7 +751,7 @@ inline OIIO_HOSTDEVICE float sign (float x)
 
 
 template <typename IN_TYPE, typename OUT_TYPE>
-OIIO_FORCEINLINE OIIO_HOSTDEVICE OUT_TYPE bit_cast (const IN_TYPE in) {
+OIIO_FORCEINLINE OIIO_HOSTDEVICE OUT_TYPE bit_cast (const IN_TYPE& in) {
     // NOTE: this is the only standards compliant way of doing this type of casting,
     // luckily the compilers we care about know how to optimize away this idiom.
     static_assert(sizeof(IN_TYPE) == sizeof(OUT_TYPE),
@@ -731,28 +768,28 @@ OIIO_FORCEINLINE OIIO_HOSTDEVICE OUT_TYPE bit_cast (const IN_TYPE in) {
     // about the memcpy in the general case.
     // FIXME: The intrinsics are not in clang <= 9 nor gcc <= 9.1. Check
     // future releases.
-    template<> OIIO_FORCEINLINE uint32_t bit_cast<float, uint32_t> (const float val) {
+    template<> OIIO_FORCEINLINE uint32_t bit_cast<float, uint32_t>(const float& val) {
           return static_cast<uint32_t>(_castf32_u32(val));
     }
-    template<> OIIO_FORCEINLINE int32_t bit_cast<float, int32_t> (const float val) {
+    template<> OIIO_FORCEINLINE int32_t bit_cast<float, int32_t>(const float& val) {
           return static_cast<int32_t>(_castf32_u32(val));
     }
-    template<> OIIO_FORCEINLINE float bit_cast<uint32_t, float> (const uint32_t val) {
+    template<> OIIO_FORCEINLINE float bit_cast<uint32_t, float>(const uint32_t& val) {
           return _castu32_f32(val);
     }
-    template<> OIIO_FORCEINLINE float bit_cast<int32_t, float> (const int32_t val) {
+    template<> OIIO_FORCEINLINE float bit_cast<int32_t, float>(const int32_t& val) {
           return _castu32_f32(val);
     }
-    template<> OIIO_FORCEINLINE uint64_t bit_cast<double, uint64_t> (const double val) {
+    template<> OIIO_FORCEINLINE uint64_t bit_cast<double, uint64_t>(const double& val) {
           return static_cast<uint64_t>(_castf64_u64(val));
     }
-    template<> OIIO_FORCEINLINE int64_t bit_cast<double, int64_t> (const double val) {
+    template<> OIIO_FORCEINLINE int64_t bit_cast<double, int64_t>(const double& val) {
           return static_cast<int64_t>(_castf64_u64(val));
     }
-    template<> OIIO_FORCEINLINE double bit_cast<uint64_t, double> (const uint64_t val) {
+    template<> OIIO_FORCEINLINE double bit_cast<uint64_t, double>(const uint64_t& val) {
           return _castu64_f64(val);
     }
-    template<> OIIO_FORCEINLINE double bit_cast<int64_t, double> (const int64_t val) {
+    template<> OIIO_FORCEINLINE double bit_cast<int64_t, double>(const int64_t& val) {
           return _castu64_f64(val);
     }
 #endif
@@ -790,6 +827,82 @@ swap_endian (T *f, int len=1)
 }
 
 
+#if (OIIO_GNUC_VERSION || OIIO_CLANG_VERSION || OIIO_APPLE_CLANG_VERSION || OIIO_INTEL_COMPILER_VERSION) && !defined(__CUDACC__)
+// CPU gcc and compatible can use these intrinsics, 8-15x faster
+
+template<> inline void swap_endian(uint16_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap16(f[i]);
+}
+
+template<> inline void swap_endian(uint32_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap32(f[i]);
+}
+
+template<> inline void swap_endian(uint64_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap64(f[i]);
+}
+
+template<> inline void swap_endian(int16_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap16(f[i]);
+}
+
+template<> inline void swap_endian(int32_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap32(f[i]);
+}
+
+template<> inline void swap_endian(int64_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap64(f[i]);
+}
+
+template<> inline void swap_endian(float* f, int len) {
+    swap_endian((uint32_t*)f, len);
+}
+
+template<> inline void swap_endian(double* f, int len) {
+    swap_endian((uint64_t*)f, len);
+}
+
+#elif defined(_MSC_VER) && !defined(__CUDACC__)
+// CPU MSVS can use these intrinsics
+
+template<> inline void swap_endian(uint16_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_ushort(f[i]);
+}
+
+template<> inline void swap_endian(uint32_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_ulong(f[i]);
+}
+
+template<> inline void swap_endian(uint64_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_uint64(f[i]);
+}
+
+template<> inline void swap_endian(int16_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_ushort(f[i]);
+}
+
+template<> inline void swap_endian(int32_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_ulong(f[i]);
+}
+
+template<> inline void swap_endian(int64_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_uint64(f[i]);
+}
+#endif
+
+
 
 // big_enough_float<T>::float_t is a floating-point type big enough to
 // handle the range and precision of a <T>. It's a float, unless T is big.
@@ -897,7 +1010,7 @@ void convert_type (const S *src, D *dst, size_t n, D _min, D _max)
 template<>
 inline void convert_type<uint8_t,float> (const uint8_t *src,
                                          float *dst, size_t n,
-                                         float _min, float _max)
+                                         float /*_min*/, float /*_max*/)
 {
     float scale (1.0f/std::numeric_limits<uint8_t>::max());
     simd::vfloat4 scale_simd (scale);
@@ -915,7 +1028,7 @@ inline void convert_type<uint8_t,float> (const uint8_t *src,
 template<>
 inline void convert_type<uint16_t,float> (const uint16_t *src,
                                           float *dst, size_t n,
-                                          float _min, float _max)
+                                          float /*_min*/, float /*_max*/)
 {
     float scale (1.0f/std::numeric_limits<uint16_t>::max());
     simd::vfloat4 scale_simd (scale);
@@ -929,11 +1042,11 @@ inline void convert_type<uint16_t,float> (const uint16_t *src,
 }
 
 
-#ifdef _HALF_H_
+#if defined(_HALF_H_) || defined(IMATH_HALF_H_)
 template<>
 inline void convert_type<half,float> (const half *src,
                                       float *dst, size_t n,
-                                      float _min, float _max)
+                                      float /*_min*/, float /*_max*/)
 {
 #if OIIO_SIMD >= 8 && OIIO_F16C_ENABLED
     // If f16c ops are enabled, it's worth doing this by 8's
@@ -958,7 +1071,7 @@ inline void convert_type<half,float> (const half *src,
 template<>
 inline void
 convert_type<float,uint16_t> (const float *src, uint16_t *dst, size_t n,
-                              uint16_t _min, uint16_t _max)
+                              uint16_t /*_min*/, uint16_t /*_max*/)
 {
     float min = std::numeric_limits<uint16_t>::min();
     float max = std::numeric_limits<uint16_t>::max();
@@ -979,7 +1092,7 @@ convert_type<float,uint16_t> (const float *src, uint16_t *dst, size_t n,
 template<>
 inline void
 convert_type<float,uint8_t> (const float *src, uint8_t *dst, size_t n,
-                             uint8_t _min, uint8_t _max)
+                             uint8_t /*_min*/, uint8_t /*_max*/)
 {
     float min = std::numeric_limits<uint8_t>::min();
     float max = std::numeric_limits<uint8_t>::max();
@@ -997,11 +1110,11 @@ convert_type<float,uint8_t> (const float *src, uint8_t *dst, size_t n,
 }
 
 
-#ifdef _HALF_H_
+#if defined(_HALF_H_) || defined(IMATH_HALF_H_)
 template<>
 inline void
 convert_type<float,half> (const float *src, half *dst, size_t n,
-                          half _min, half _max)
+                          half /*_min*/, half /*_max*/)
 {
 #if OIIO_SIMD >= 8 && OIIO_F16C_ENABLED
     // If f16c ops are enabled, it's worth doing this by 8's
@@ -1104,7 +1217,7 @@ bit_range_convert(unsigned int in, unsigned int FROM_BITS, unsigned int TO_BITS)
 
 
 /// Append the `n` LSB bits of `val` into a bit sting `T out[]`, where the
-/// `filled` MSB bits of `*out` are already filled in. Incremennt `out` and
+/// `filled` MSB bits of `*out` are already filled in. Increment `out` and
 /// adjust `filled` as required. Type `T` should be uint8_t, uint16_t, or
 /// uint32_t.
 template<typename T>
@@ -1792,7 +1905,7 @@ OIIO_FORCEINLINE OIIO_HOSTDEVICE float fast_atan2 (float y, float x) {
 template<typename T>
 OIIO_FORCEINLINE OIIO_HOSTDEVICE T fast_log2 (const T& xval) {
     using namespace simd;
-    typedef typename T::int_t intN;
+    typedef typename T::vint_t intN;
     // See float fast_log2 for explanations
     T x = clamp (xval, T(std::numeric_limits<float>::min()), T(std::numeric_limits<float>::max()));
     intN bits = bitcast_to_int(x);
@@ -1900,7 +2013,7 @@ OIIO_FORCEINLINE OIIO_HOSTDEVICE float fast_log1p (float x) {
 template<typename T>
 OIIO_FORCEINLINE OIIO_HOSTDEVICE T fast_exp2 (const T& xval) {
     using namespace simd;
-    typedef typename T::int_t intN;
+    typedef typename T::vint_t intN;
 #if OIIO_SIMD_SSE
     // See float specialization for explanations
     T x = clamp (xval, T(-126.0f), T(126.0f));
@@ -1932,7 +2045,7 @@ OIIO_FORCEINLINE OIIO_HOSTDEVICE T fast_exp2 (const T& xval) {
 
 template<>
 OIIO_FORCEINLINE OIIO_HOSTDEVICE float fast_exp2 (const float& xval) {
-#if OIIO_NON_INTEL_CLANG && OIIO_FMATH_SIMD_FRIENDLY
+#if OIIO_ANY_CLANG && !OIIO_INTEL_LLVM_COMPILER && OIIO_FMATH_SIMD_FRIENDLY
     // Clang was unhappy using the bitcast/memcpy/reinter_cast/union inside
     // an explicit SIMD loop, so revert to calling the standard version.
     return std::exp2(xval);
@@ -2230,7 +2343,7 @@ T invert (Func &func, T y, T xmin=0.0, T xmax=1.0,
         *brack = bracketed;
     if (! bracketed) {
         // If our bounds don't bracket the zero, just give up, and
-        // return the approprate "edge" of the interval
+        // return the appropriate "edge" of the interval
         return ((y < vmin) == increasing) ? xmin : xmax;
     }
     if (fabs(v0-v1) < eps)   // already close enough
